#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import hashlib
import subprocess
import xml.dom.minidom
from .Tools import XMLTools
from .TOC import TOCElement, TOCDirective
from .Slide import RenderSlideDirective
from .Acronyms import AcronymDirective
from .Exceptions import MalformedXMLInputException

class Presentation():
	_NAMESPACES = {
		"https://github.com/johndoe31415/pyradium":		"s",
	}

	def __init__(self, meta, content, sources):
		self._meta = meta
		self._content = content
		self._sources = sources
		self._validate_metadata()

	@property
	def meta(self):
		return self._meta

	@property
	def content(self):
		return self._content

	@property
	def sources(self):
		return self._sources

	@classmethod
	def load_from_file(cls, filename, rendering_parameters = None):
		try:
			dom = xml.dom.minidom.parse(filename)
		except xml.parsers.expat.ExpatError as e:
			raise MalformedXMLInputException("Cannot parse %s: %s" % (filename, str(e)))
		cls._NAMESPACES.update(XMLTools.normalize_ns(dom.documentElement, cls._NAMESPACES))
		meta = None
		content = [ ]
		sources = [ filename ]
		presentation = XMLTools.child_tagname(dom, "presentation")
		if presentation is None:
			raise MalformedXMLInputException("No 'presentation' node is present as top node of the XML input document.")
		for child in presentation.childNodes:
			if child.nodeType != child.ELEMENT_NODE:
				continue

			if child.tagName == "meta":
				meta = XMLTools.xml_to_dict(XMLTools.child_tagname(dom, ("presentation", "meta")))
			elif child.tagName == "slide":
				if not XMLTools.get_bool_attr(child, "hide"):
					content.append(RenderSlideDirective(child))
			elif child.tagName == "include":
				sub_presentation_filename = rendering_parameters.include_dirs.lookup(child.getAttribute("src"))
				sub_presentation = cls.load_from_file(sub_presentation_filename)
				content += sub_presentation.content
				sources += sub_presentation.sources
			elif child.tagName in [ "chapter", "section", "subsection" ]:
				toc_element = TOCElement(child.tagName)
				toc_directive = TOCDirective(toc_element, XMLTools.inner_text(child))
				content.append(toc_directive)
			elif child.tagName == "acronyms":
				content.append(AcronymDirective(child))
			else:
				print("Warning: Ignored unknown tag '%s'." % (child.tagName))
		return cls(meta, content, sources)

	def _validate_metadata(self):
		if self.meta is not None:
			if not isinstance(self.meta.get("schedule", { }), dict):
				raise MalformedXMLInputException("The 'meta' child node 'schedule' must not contain text, but two child nodes (<total-presentation-time> and <pause-minutes>).")

	def _determine_sha256(self, filename):
		with open(filename, "rb") as f:
			return hashlib.sha256(f.read()).hexdigest()

	def _determine_git_commit(self, filename):
		try:
			return subprocess.check_output([ "git", "rev-list", "-1", "--all", filename ], stderr = subprocess.DEVNULL).decode().rstrip("\n")
		except subprocess.CalledProcessError:
			return None

	def _determine_version(self, filename):
		return {
			"sha256":	self._determine_sha256(filename),
			"git":		self._determine_git_commit(filename),
		}

	@property
	def version_information(self):
		return { filename: self._determine_version(filename) for filename in self.sources }

	def __iter__(self):
		return iter(self._content)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import shutil
import xml.dom.minidom
from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.CmdlineParser import CmdlineParser
from pyradium.Exceptions import FailedToLookupFileException, MalformedXMLInputException
from pyradium.CmdlineEscape import CmdlineEscape

@XMLHookRegistry.register_hook
class ExecHook(BaseHook):
	_TAG_NAME = "exec"

	@classmethod
	def handle(cls, rendered_presentation, node):
		if not node.hasAttribute("cmd"):
			raise MalformedXMLInputException("An s:exec node must have a 'cmd' attribute.")

		cmd_str = node.getAttribute("cmd")
		cmd = CmdlineParser().parse(cmd_str)
		if not cmd[0].startswith("/"):
			# Not an absolute path given, look up path of executable
			try:
				full_executable_path = rendered_presentation.renderer.lookup_include(cmd[0])
			except FailedToLookupFileException as e:
				# Try if this is a path command
				full_executable_path = shutil.which(cmd[0])
				if full_executable_path is None:
					raise FailedToLookupFileException("Could lookup executable \"%s\" neither in include dirs (%s) not in $PATH." % (cmd[0], str(e)))

		cmd[0] = full_executable_path
		properties = {
			"cmd":		cmd,
			"cache":	XMLTools.get_bool_attr(node, "cache", default_value = True),
		}

		exec_renderer = rendered_presentation.renderer.get_custom_renderer("exec")
		result = exec_renderer.render(properties)

		xml_data = result.data["stdout"]
		try:
			doc = xml.dom.minidom.parseString(xml_data)
		except xml.parsers.expat.ExpatError as e:
			raise MalformedXMLInputException("Ouptut of script execution (%s) returned invalid XML: %s" % (CmdlineEscape().cmdline(cmd), str(e)))

		root = doc.firstChild
		return list(root.childNodes)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import contextlib
import hashlib
import datetime
import collections
import json
from .ExtendedJSONEncoder import ExtendedJSONEncoder

RenderedResult = collections.namedtuple("RenderedResult", [ "key", "keyhash", "from_cache", "data" ])

class BaseRenderer():
	RendererResult = collections.namedtuple("RendererResult", [ "key", "data" ])

	@property
	def name(self):
		raise NotImplementedError(__class__.__name__)

	@property
	def properties(self):
		return { "version": 0 }

	def rendering_key(self, property_dict):
		return None

	def render(self, property_dict):
		raise NotImplementedError(__class__.__name__)

class RendererCache():
	def __init__(self, renderer):
		self._renderer = renderer
		self._directory = os.path.expanduser("~/.cache/pyradium/" + self._renderer.name + "/")
		with contextlib.suppress(FileExistsError):
			os.makedirs(self._directory)

	@staticmethod
	def _hash_key(key):
		binkey = ExtendedJSONEncoder.dumps(key, minify = True, sort_keys = True).encode("utf-8")
		keyhash = hashlib.md5(binkey).hexdigest()
		return keyhash

	def _retrieve(self, keyhash):
		filename = self._directory + keyhash + ".json"
		try:
			with open(filename) as f:
				file_representation = ExtendedJSONEncoder.load(f)
		except (FileNotFoundError, json.decoder.JSONDecodeError):
			return None

		return RenderedResult(key = file_representation["key"], keyhash = keyhash, from_cache = True, data = file_representation["object"])

	def _store(self, key, keyhash, object_data):
		file_representation = {
			"key":		key,
			"meta": {
				"rendered":	datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
				"keyhash":	keyhash,
			},
			"object": object_data,
		}

		filename = self._directory + keyhash + ".json"
		with open(filename, "w") as f:
			ExtendedJSONEncoder.dump(file_representation, f, minify = True)

	def render(self, property_dict):
		key = {
			"name":						self._renderer.name,
			"renderer_properties":		self._renderer.properties,
			"object_properties":		property_dict,
			"additional_key":			self._renderer.rendering_key(property_dict),
		}
		attempt_cache = property_dict.get("cache", True)

		keyhash = self._hash_key(key)
		cached_object = self._retrieve(keyhash)
		if attempt_cache and (cached_object is not None):
			return cached_object
		else:
			object_data = self._renderer.render(property_dict)
			if attempt_cache:
				self._store(key, keyhash, object_data)
			return RenderedResult(key = key, keyhash = keyhash, from_cache = False, data = object_data)

if __name__ == "__main__":
	class DebugRenderer(BaseRenderer):
		@property
		def name(self):
			return "debug"

		@property
		def properties(self):
			return { "version": 1 }

		def render(self, property_dict):
			return {
				"text":		property_dict["letter"] * property_dict["count"],
				"bytes":	b"foobar" * 1000,
			}

	cache = RendererCache(DebugRenderer())
	print("Result = ", cache.render({
		"letter":	"Q",
		"count":	10,
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import unittest
from pyradium.Schedule import TimeSpecificationType, TimeSpecification, TimeRange, TimeRanges, PresentationSchedule
from pyradium.Exceptions import TimeSpecificationError

class TimeSpecificationTests(unittest.TestCase):
	def test_must_supply_value(self):
		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse()

	def test_may_not_supply_two_values(self):
		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(abs_string = "1 min", rel_string = "1")

	def test_rel_value(self):
		tspec = TimeSpecification.parse(rel_string = "1.234")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Relative)
		self.assertAlmostEqual(tspec.value, 1.234)

	def test_rel_value_noparse(self):
		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(rel_string = "")

		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(rel_string = "ABC")

	def test_abs_value_secs(self):
		tspec = TimeSpecification.parse(abs_string = "10 sec")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 10)

		tspec = TimeSpecification.parse(abs_string = "123s")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 123)

	def test_abs_value_min(self):
		tspec = TimeSpecification.parse(abs_string = "1min")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 60)

		tspec = TimeSpecification.parse(abs_string = "1.5m")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 90)

	def test_abs_value_colon(self):
		tspec = TimeSpecification.parse(abs_string = "2:30")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 150)

		tspec = TimeSpecification.parse(abs_string = "0:12")
		self.assertEqual(tspec.spec_type, TimeSpecificationType.Absolute)
		self.assertAlmostEqual(tspec.value, 12)

	def test_abs_value_noparse(self):
		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(abs_string = "")

		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(abs_string = ":12")

		with self.assertRaises(TimeSpecificationError):
			TimeSpecification.parse(abs_string = "1.234:30 min")

	def test_timerange_simple(self):
		tr = TimeRange.parse("11:12 - 22:23")
		self.assertEqual(tr.range_begin, (11 * 60) + 12)
		self.assertEqual(tr.range_end, (22 * 60) + 23)
		self.assertEqual(tr.begin_time, (11, 12))
		self.assertEqual(tr.end_time, (22, 23))
		self.assertEqual(tr.duration_mins, 48 + (60 * 10) + 23)

	def test_timerange_crossover(self):
		tr = TimeRange.parse(" 23:45-0:15 ")
		self.assertEqual(tr.range_begin, (23 * 60) + 45)
		self.assertEqual(tr.range_end, 1440 + (0 * 60) + 15)
		self.assertEqual(tr.begin_time, (23, 45))
		self.assertEqual(tr.end_time, (0, 15))
		self.assertEqual(tr.duration_mins, 30)

	def test_timerange_noparse(self):
		with self.assertRaises(TimeSpecificationError):
			TimeRange.parse("0 : 10 - 10:12")
		with self.assertRaises(TimeSpecificationError):
			TimeRange.parse(" 23:23 - 24:12")
		with self.assertRaises(TimeSpecificationError):
			TimeRange.parse(" 23:23 - 24:12")
		with self.assertRaises(TimeSpecificationError):
			TimeRange.parse(" 23:23 - 22:x2")

	def test_timeranges_simple(self):
		trs = TimeRanges.parse("10:00-10:15 10:20-10:30")
		self.assertEqual(trs.duration_mins, 15 + 10)

	def test_presentation_schedule_simple(self):
		ps = PresentationSchedule(active_presentation_time_minutes = 45)
		ps.set_slide_no(1, TimeSpecification.parse(abs_string = "2 min"))
		ps.set_slide_no(5, TimeSpecification.parse(rel_string = "2"))
		ps.have_slide(15)
#		for x in ps:
#			print(x)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import subprocess
from pyradium.Tools import HashTools
from pyradium.RendererCache import BaseRenderer
from pyradium.Exceptions import FailedToExecuteSubprocessException

class ExecRenderer(BaseRenderer):
	def __init__(self):
		super().__init__()

	@property
	def name(self):
		return "exec"

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def rendering_key(self, property_dict):
		cmd = property_dict["cmd"]
		return {
			"srchash":		HashTools.hash_file(cmd[0]),
		}

	def render(self, property_dict):
		cmd = property_dict["cmd"]
		try:
			proc = subprocess.run(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		except PermissionError as e:
			raise FailedToExecuteSubprocessException("Could not execute '%s' (%s)." % (str(cmd), str(e)))

		if proc.returncode != 0:
			raise FailedToExecuteSubprocessException("Could not execute '%s': returncode %d." % (str(cmd), proc.returncode))

		result = {
			"cmd":			cmd,
			"stdout":		proc.stdout,
			"stderr":		proc.stderr,
		}
		return result

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(ExecRenderer())
	print(renderer.render({
		"cmd":	[ "/bin/ls", "-l", "/" ],
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
from .BaseDirective import BaseDirective

class TOCElement(enum.Enum):
	Chapter = "chapter"
	Section = "section"
	SubSection = "subsection"

class TOCDirective(BaseDirective):
	_TOC_LEVEL = {
		TOCElement.Chapter:		0,
		TOCElement.Section:		1,
		TOCElement.SubSection:	2,
	}

	def __init__(self, toc_element, value):
		assert(isinstance(toc_element, TOCElement))
		self._toc_element = toc_element
		self._value = value

	@property
	def toc_element(self):
		return self._toc_element

	@property
	def value(self):
		return self._value

	def render(self, rendered_presentation):
		level = self._TOC_LEVEL[self.toc_element]
		rendered_presentation.toc.new_heading(level, self.value)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import ReplacementHook, XMLHookRegistry

@XMLHookRegistry.register_hook
class EmoHook(ReplacementHook):
	_TAG_NAME = "emo"
	_REPLACEMENTS = {
		";)":	"😉",
		":)":	"🙂",
		":))":	"😀",
		":(":	"🙁",
		":((":	"☹️",
		";(":	"😢",
		";((":	"😭",
		":D":	"😁",
		":?":	"🤔",
		":|":	"😐",
		":/":	"😕",
		":o":	"😲",
		":O":	"😮",
	}



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import tempfile
import subprocess
from pyradium.RendererCache import BaseRenderer

_TEX_TEMPLATE = r"""
\documentclass[preview,border=1mm,varwidth=true]{standalone}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{lmodern}
\usepackage{amsmath}
\usepackage{amssymb}
\begin{document}
%(content)s
\end{document}
"""

class LatexFormulaRenderer(BaseRenderer):
	def __init__(self, rendering_dpi = 600):
		super().__init__()
		self._rendering_dpi = rendering_dpi

	@property
	def name(self):
		return "latex"

	@property
	def properties(self):
		return {
			"version":			1,
			"rendering_dpi":	self._rendering_dpi,
		}

	def _get_image_info(self, png_filename, xoffset):
		# Need 2 pixel wide sample so that ImageMagick does not return an empty image
		crop_info = json.loads(subprocess.check_output([ "convert", "%s[2x+%d+0]" % (png_filename, xoffset), "-trim", "json:-" ]))

		image_width = crop_info[0]["image"]["pageGeometry"]["width"]
		image_height = crop_info[0]["image"]["pageGeometry"]["height"]
		upper_baseline_y_from_top = crop_info[0]["image"]["pageGeometry"]["y"]
		lower_baseline_y_from_top = upper_baseline_y_from_top + crop_info[0]["image"]["geometry"]["height"]
		baseline_y_from_top = round((upper_baseline_y_from_top + lower_baseline_y_from_top) / 2)
		baseline_y_from_bottom = image_height - baseline_y_from_top

		image_info = {
			"width":		image_width,
			"height":		image_height,
			"baseline":		baseline_y_from_bottom,
		}
		return image_info

	def render(self, property_dict):
		with tempfile.TemporaryDirectory(prefix = "pyradium_formula_") as tex_dir:
			# Formula to PDF first using pdflatex
			tex_filename = tex_dir + "/formula.tex"
			pdf_filename = tex_dir + "/formula.pdf"
			png_filename = tex_dir + "/formula.png"

			# Crop 3mm off the left side (1mm baseline bar + 2mm space)
			baseline = r"\rule{1mm}{1pt} \hspace{2mm}"
			left_crop_pixel = round((3 / 25.4) * self._rendering_dpi)
			left_crop_pixel_safe = round((2 / 25.4) * self._rendering_dpi)
			eval_baseline_at_x = round((0.5 / 25.4) * self._rendering_dpi)

			if property_dict.get("long", False):
				content = r"\[" + baseline + property_dict["formula"] + r" \]"
			else:
				content = r"$" + baseline + property_dict["formula"] + r"$"
			with open(tex_filename, "w") as tex_file:
				tex_file.write(_TEX_TEMPLATE % { "content": content })
			subprocess.check_call([ "pdflatex", "-output-directory=%s" % (tex_dir), tex_filename ])

			# Then render the PDF to PNG
			subprocess.check_call([ "convert", "-define", "profile:skip=ICC", "-density", str(self._rendering_dpi), "-trim", "+repage", pdf_filename, png_filename ])

#			print("Crop on left side: %d (choosing %d to be on safe side); evaluating baseline at x = %d; filename %s" % (left_crop_pixel, left_crop_pixel_safe, eval_baseline_at_x, png_filename))
			image_info = self._get_image_info(png_filename, eval_baseline_at_x)

#			print("Determined baseline y-offset at %d pixels (measured from bottom)" % (image_info["baseline"]))

			# Then crop the image finally
			png_data = subprocess.check_output([ "convert", "-crop", "+%d+0" % (left_crop_pixel_safe), "-trim", "+repage", png_filename, "png:-" ])

			# Return an image object
			image = {
				"png_data":	png_data,
				"info":		image_info,
			}
			return image

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(LatexFormulaRenderer())
	print(renderer.render({ "formula": "y^2 = x^3 + ax + b" }))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import json
import contextlib
from .GenericTOC import GenericTOC
from .OrderedSet import OrderedSet
from .Schedule import PresentationSchedule, TimeRange

class RenderedPresentation():
	def __init__(self, renderer, deploy_directory):
		self._renderer = renderer
		self._deploy_directory = deploy_directory
		self._rendered_slides = [ ]
		self._css = { }
		self._js = OrderedSet()
		self._toc = GenericTOC()
		self._frozen_toc = None
		self._added_files = set()
		self._current_slide_number = 0
		self._total_slide_count = 0
		self._uid = 0
		self._features = set()

		time_range = self._renderer.presentation.meta.get("schedule", { }).get("total-presentation-time")
		if time_range is None:
			presentation_length_mins = None
		else:
			presentation_length_mins = TimeRange.parse(time_range).duration_mins
		self._schedule = PresentationSchedule(presentation_length_mins)

	@property
	def next_unique_id(self):
		self._uid += 1
		return "uid_%x" % (self._uid)

	@property
	def current_slide_number(self):
		return self._current_slide_number

	@property
	def total_slide_count(self):
		return self._total_slide_count

	def advance_slide(self):
		self._current_slide_number += 1
		self._total_slide_count = max(self._total_slide_count, self._current_slide_number)
		self._toc.at_page(self.current_slide_number)
		self.schedule.have_slide(self.current_slide_number)

	def finalize_toc(self):
		self._frozen_toc = self._toc.finalize()
		self._toc = GenericTOC()
		self._current_slide_number = 0

	@property
	def renderer(self):
		return self._renderer

	@property
	def rendered_slides(self):
		return iter(self._rendered_slides)

	@property
	def js(self):
		return iter(self._js)

	@property
	def css(self):
		for (filename, order) in sorted(self._css.items(), key = lambda x: (x[1], x[0])):
			yield filename

	@property
	def schedule(self):
		return self._schedule

	def add_css(self, filename, order = None, target_directory = "/"):
		assert(target_directory.startswith("/"))
		assert(target_directory.endswith("/"))
		if order is None:
			order = len(self._css)
		key = target_directory[1:] + filename
		if key not in self._css:
			self._css[key] = order

	def add_js(self, filename):
		return self._js.add(filename)

	@property
	def toc(self):
		return self._toc

	@property
	def frozen_toc(self):
		return self._frozen_toc

	@property
	def features(self):
		return self._features

	def add_feature(self, feature):
		self._features.add(feature)

	def has_feature(self, feature):
		return feature in self.features

	def append_slide(self, rendered_slide):
		self._rendered_slides.append(rendered_slide)

	def add_file(self, destination_relpath, content, target_directory = "/"):
		assert(target_directory.startswith("/"))
		assert(target_directory.endswith("/"))
		if destination_relpath in self._added_files:
			return
		self._added_files.add(destination_relpath)
		filename = self._deploy_directory + target_directory + destination_relpath
		dirname = os.path.dirname(filename)
		with contextlib.suppress(FileExistsError):
			os.makedirs(dirname)
		with open(filename, "w" if isinstance(content, str) else "wb") as f:
			f.write(content)

	def copy_file(self, rel_filename, target_directory = "/"):
		if rel_filename in self._added_files:
			return
		source_filename = self.renderer.lookup_template_file(rel_filename)
		with open(source_filename, "rb") as f:
			self.add_file(rel_filename, f.read(), target_directory)

	def handle_dependencies(self, dependencies):
		if dependencies is None:
			return
		for rel_filename in dependencies.get("static", [ ]):
			self.copy_file(rel_filename, target_directory = "/template/")
		for rel_filename in dependencies.get("css", [ ]):
			if isinstance(rel_filename, str):
				self.copy_file(rel_filename, target_directory = "/template/")
				self.add_css(rel_filename, target_directory = "/template/")
			else:
				if rel_filename.get("render"):
					rendered_css = self.renderer.render_file(rel_filename["name"], rendered_presentation = self)
					self.add_file(rel_filename["name"], rendered_css, target_directory = "/template/")
				else:
					self.copy_file(rel_filename["name"], target_directory = "/template/")
				self.add_css(rel_filename["name"], target_directory = "/template/", order = rel_filename.get("order"))

	@property
	def meta(self):
		return {
			"total_presentation_time":	self.renderer.presentation.meta.get("schedule", { }).get("total-presentation-time"),
			"pause_minutes":			self.renderer.presentation.meta.get("schedule", { }).get("pause-minutes"),
		}

	@property
	def meta_json(self):
		return json.dumps(self.meta)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
from .Enums import PresentationMode
from .FileLookup import FileLookup

class RenderingParameters():
	def __init__(self, template_style = "default", honor_pauses = True, collapse_animation = False, presentation_mode = PresentationMode.Interactive, extra_template_dirs = None, include_dirs = None, index_filename = "index.html", geometry = (1280, 720), image_max_dimension = 1920, presentation_features = None):
		self._template_style = template_style
		self._honor_pauses = honor_pauses
		self._collapse_animation = collapse_animation
		self._presentation_mode = presentation_mode
		template_dirs = [ os.path.expanduser("~/.config/pyradium/templates"), os.path.dirname(os.path.realpath(__file__)) + "/templates" ]
		if extra_template_dirs is not None:
			template_dirs += extra_template_dirs
		self._template_dirs = FileLookup(template_dirs)
		if include_dirs is None:
			self._include_dirs = FileLookup()
		else:
			self._include_dirs = FileLookup(include_dirs)
		self._index_filename = index_filename
		self._geometry = geometry
		self._image_max_dimension = image_max_dimension
		self._presentation_features = set(presentation_features) if (presentation_features is not None) else set()

	@property
	def template_style(self):
		return self._template_style

	@property
	def honor_pauses(self):
		return self._honor_pauses

	@property
	def collapse_animation(self):
		return self._collapse_animation

	@property
	def presentation_mode(self):
		return self._presentation_mode

	@property
	def template_dirs(self):
		return self._template_dirs

	@property
	def include_dirs(self):
		return self._include_dirs

	@property
	def index_filename(self):
		return self._index_filename

	@property
	def geometry_x(self):
		return self._geometry[0]

	@property
	def geometry_y(self):
		return self._geometry[1]

	@property
	def image_max_dimension(self):
		return self._image_max_dimension

	@property
	def presentation_features(self):
		return iter(self._presentation_features)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import xml.dom.minidom
import pygments
from pyradium.xmlhooks.XMLHookRegistry import InnerTextHook, XMLHookRegistry

@XMLHookRegistry.register_hook
class CodeHook(InnerTextHook):
	_TAG_NAME = "code"

	@classmethod
	def handle_text(cls, text, rendered_presentation, node):
		lexer = pygments.lexers.get_lexer_by_name(node.getAttribute("lang"))
		highlighted_code = pygments.highlight(text, lexer, pygments.formatters.HtmlFormatter(cssclass = "code_highlight"))

		replacement_node = xml.dom.minidom.parseString(highlighted_code).firstChild
		rendered_presentation.add_feature("pygments")
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import mako.lookup
import pyradium
from pyradium.Controller import ControllerManager
from pyradium.renderer.LatexFormulaRenderer import LatexFormulaRenderer
from pyradium.renderer.ImageRenderer import ImageRenderer
from pyradium.renderer.ExecRenderer import ExecRenderer
from .Acronyms import Acronyms
from .RendererCache import RendererCache
from .RenderedPresentation import RenderedPresentation
from .Enums import PresentationMode
from .Exceptions import TemplateErrorException
from .Slide import RenderSlideDirective

class Renderer():
	def __init__(self, presentation, rendering_params):
		self._rendered_slides = [ ]
		self._presentation = presentation
		self._rendering_params = rendering_params
		self._custom_renderers = {
			"latex":	RendererCache(LatexFormulaRenderer()),
			"img":		RendererCache(ImageRenderer()),
			"exec":		RendererCache(ExecRenderer()),
			"acronym":	Acronyms(),
		}
		self._lookup = mako.lookup.TemplateLookup(list(self._get_mako_lookup_directories()), strict_undefined = True, input_encoding = "utf-8", default_filters = [ "h" ])
		with open(self.lookup_styled_template_file("configuration.json")) as f:
			self._template_config = json.load(f)
		self._ctrlr_mgr = ControllerManager(self)

	@property
	def presentation(self):
		return self._presentation

	def _get_mako_lookup_directories(self):
		for dirname in self._rendering_params.template_dirs:
			yield dirname
			yield dirname + "/" + self._rendering_params.template_style

	@property
	def rendering_params(self):
		return self._rendering_params

	@property
	def template_config(self):
		return self._template_config

	@property
	def controllers(self):
		return self._ctrlr_mgr

	def get_custom_renderer(self, name):
		return self._custom_renderers[name]

	def lookup_template_file(self, filename):
		return self._rendering_params.template_dirs.lookup(filename)

	def lookup_styled_template_file(self, filename):
		return self._rendering_params.template_dirs.lookup(self._rendering_params.template_style + "/" + filename)

	def lookup_include(self, filename):
		return self._rendering_params.include_dirs.lookup(filename)

	def _determine_slide_types(self):
		slide_types = set()
		for directive in self._presentation:
			if isinstance(directive, RenderSlideDirective):
				slide_types.add(directive.slide_type)
		return slide_types

	def _compute_renderable_slides(self, rendered_presentation):
		renderable_slides = [ ]
		for directive in self._presentation:
			generator = directive.render(rendered_presentation)
			if generator is not None:
				renderable_slides += generator
		return renderable_slides

	def render_file(self, template_filename, rendered_presentation = None, additional_template_args = None):
		def _template_error(text):
			raise TemplateErrorException(text)

		template_args = {
			"pyradium_version":			pyradium.VERSION,
			"renderer":					self,
			"presentation":				self._presentation,
			"template_error":			_template_error,
		}
		if rendered_presentation is not None:
			template_args["rendered_presentation"] = rendered_presentation
		if additional_template_args is not None:
			template_args.update(additional_template_args)

		template = self._lookup.get_template(template_filename)
		result = template.render(**template_args)
		return result

	def render(self, deploy_directory):
		rendered_presentation = RenderedPresentation(self, deploy_directory = deploy_directory)

		for feature in self.rendering_params.presentation_features:
			rendered_presentation.add_feature(feature.value)
		if self.rendering_params.presentation_mode == PresentationMode.Interactive:
			rendered_presentation.add_feature("interactive")

		# Run it first to build the initial TOC and determine feature set
		self._compute_renderable_slides(rendered_presentation)


		# Then copy the dependencies
		rendered_presentation.handle_dependencies(self._template_config.get("files"))
		for slide_type in self._determine_slide_types():
			rendered_presentation.handle_dependencies(self._template_config.get("dependencies", { }).get("slidetype", { }).get(slide_type))
		for feature in rendered_presentation.features:
			rendered_presentation.handle_dependencies(self._template_config.get("dependencies", { }).get("feature", { }).get(feature))

		# Then run it again to get the page numbers straight (e.g., the TOC
		# pages will be emitted, giving different page numbers)
		rendered_presentation.finalize_toc()
		self._compute_renderable_slides(rendered_presentation)
		rendered_presentation.finalize_toc()

		# Compute the schedule
		rendered_presentation.schedule.compute()

		for renderable_slide in self._compute_renderable_slides(rendered_presentation):
			additional_template_args = {
				"slide":			renderable_slide,
			}
			template_filename = "slide_%s.html" % (renderable_slide.slide_type)
			rendered_slide = self.render_file(template_filename, rendered_presentation = rendered_presentation, additional_template_args = additional_template_args)
			rendered_presentation.append_slide(rendered_slide)

		rendered_index = self.render_file("base/index.html", rendered_presentation = rendered_presentation)
		rendered_presentation.add_file(self.rendering_params.index_filename, rendered_index)
		return rendered_presentation



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import sys
import time
import subprocess
from .BaseAction import BaseAction
from .Presentation import Presentation
from .RenderingParameters import RenderingParameters
from .Renderer import Renderer
from .Exceptions import CallingProcessException, PyRadiumException

class ActionRender(BaseAction):
	def _wait_for_change(self, renderer):
		cmd = [ "inotifywait" ]
		cmd += [ "-q", "-r" ]
		cmd += [ "--exclude", r"\..*\.sw[a-z]" ]
		cmd += [ "-e", "modify", "-e", "close_write", "-e", "create" ]

		sources = [ ]
		sources += renderer.presentation.sources
		sources += renderer.rendering_params.template_dirs
		sources += renderer.rendering_params.include_dirs
		sources += self._args.re_render_watch
		sources = [ source for source in sources if os.path.exists(source) ]
		cmd += sources
		proc = subprocess.run(cmd)
		if proc.returncode not in [ 0, 1 ]:
			raise CallingProcessException("inotifywait returned with returncode %d." % (proc.returncode))

	def run(self):
		if (not self._args.force) and os.path.exists(self._args.outdir):
			print("Refusing to overwrite: %s" % (self._args.outdir))
			return 1

		renderer = None
		render_success = True
		while True:
			try:
				rendering_parameters = RenderingParameters(
						template_style = self._args.template_style,
						honor_pauses = not self._args.remove_pauses,
						collapse_animation = self._args.collapse_animation,
						presentation_mode = self._args.presentation_mode,
						extra_template_dirs = self._args.template_dir,
						include_dirs = [ os.path.dirname(self._args.infile) or "." ] + self._args.include_dir,
						index_filename = self._args.index_filename,
						geometry = self._args.geometry,
						image_max_dimension = self._args.image_max_dimension,
						presentation_features = self._args.presentation_feature)
				presentation = Presentation.load_from_file(self._args.infile, rendering_parameters)
				renderer = Renderer(presentation, rendering_parameters)
				rendered_presentation = renderer.render(deploy_directory = self._args.outdir)
			except PyRadiumException as e:
				render_success = False
				print("Rendering failed: [%s] %s" % (e.__class__.__name__, str(e)), file = sys.stderr)
			if not self._args.re_render_loop:
				break
			if renderer is None:
				sleep_duration_secs = 5
				print("Unable to watch files for change since parsing the source was impossible; sleeping for %d seconds instead." % (sleep_duration_secs), file = sys.stderr)
				time.sleep(sleep_duration_secs)
			else:
				self._wait_for_change(renderer)
		return 0 if render_success else 1



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class CmdlineParserException(Exception): pass
class TrailingEscapeCharException(CmdlineParserException): pass
class InvalidEscapeCharException(CmdlineParserException): pass

class CmdlineParser():
	def __init__(self):
		pass

	def parse(self, cmdline):
		cmdline = cmdline.strip(" ")
		args = [ "" ]
		current_escape = None
		i = 0
		while i < len(cmdline):
			char = cmdline[i]
			if char == "\\":
				if i == len(cmdline) - 1:
					raise TrailingEscapeCharException("End of string ends with meta character.")

				i += 1
				nextchar = cmdline[i]
				if nextchar not in [ "\\", "\"", "'", " " ]:
					raise InvalidEscapeCharException("Not a valid character for escaping: %s" % (nextchar))

				args[len(args) - 1] += nextchar
			elif current_escape is None:
				if char == " ":
					if len(args[len(args) - 1]) > 0:
						args.append("")
				elif char == "\"":
					current_escape = "\""
				elif char == "'":
					current_escape = "'"
				else:
					args[len(args) - 1] += char
			elif current_escape == "\"":
				if char == "\"":
					current_escape = None
				else:
					args[len(args) - 1] += char
			elif current_escape == "'":
				if char == "'":
					current_escape = None
				else:
					args[len(args) - 1] += char
			else:
				raise NotImplementedError(current_escape)

			i += 1
		return args



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import textwrap
from pyradium.Exceptions import XMLHookRegistryException
from pyradium.Tools import XMLTools

class XMLHookRegistry():
	_HOOKS = { }
	_SPECIAL = set([ "var", "pause", "content" ])

	@classmethod
	def register_hook(cls, hook_class):
		if hook_class._TAG_NAME is None:
			raise XMLHookRegistryException("Hook name is not set in hook class: %s" % (hook_class))
		if hook_class._TAG_NAME in cls._HOOKS:
			raise XMLHookRegistryException("Duplicate hook name: %s" % (hook_class))
		cls._HOOKS[hook_class._TAG_NAME] = hook_class
		return hook_class

	@classmethod
	def _replace_text(self, text):
		text = text.replace("---", "—")
		text = text.replace("--", "–")
		return text

	@classmethod
	def mangle(cls, rendered_presentation, root_node):
		def callback(node):
			if (node.nodeType == node.ELEMENT_NODE) and (node.nodeName.startswith("s:")):
				hook_name = node.nodeName[2:]
				if hook_name not in cls._SPECIAL:
					if hook_name in cls._HOOKS:
						hook_class = cls._HOOKS[hook_name]
						replace_by = hook_class.handle(rendered_presentation, node)
						if replace_by is None:
							# Delete node
							XMLTools.remove_node(node)
						else:
							XMLTools.replace_node(node, replace_by)
					else:
						print("Warning: Unknown hook '%s' used in source document." % (hook_name))
			elif node.nodeType == node.TEXT_NODE:
				text = node.wholeText
				new_text = cls._replace_text(text)
				if text != new_text:
					node.replaceWholeText(new_text)
		XMLTools.walk(root_node, callback)

class BaseHook():
	_TAG_NAME = None

	@classmethod
	def handle(cls, rendered_presentation, node):
		raise NotImplementedError("%s.handle" % (cls.__name__))

class ReplacementHook(BaseHook):
	_REPLACEMENTS = None
	_SPAN_ATTRIBUTES = None

	@classmethod
	def handle(cls, rendered_presentation, node):
		text = XMLTools.inner_text(node).strip()
		if text not in cls._REPLACEMENTS:
			print("Warning: %s(%s) not available. Ignored the sequence." % (cls._TAG_NAME, text))
			return None
		else:
			replacement_text = cls._REPLACEMENTS[text]

			text_node = node.ownerDocument.createTextNode(replacement_text)
			if cls._SPAN_ATTRIBUTES is None:
				replacement_node = text_node
			else:
				replacement_node = node.ownerDocument.createElement("span")
				for (key, value) in cls._SPAN_ATTRIBUTES.items():
					replacement_node.setAttribute(key, value)
				replacement_node.appendChild(text_node)

			return replacement_node

class InnerTextHook(BaseHook):
	@classmethod
	def handle_text(cls, text, rendered_presentation, node):
		raise NotImplementedError("%s.handle_text" % (cls.__name__))

	@classmethod
	def handle(cls, rendered_presentation, node):
		if node.hasAttribute("src"):
			with open(rendered_presentation.renderer.lookup_include(node.getAttribute("src"))) as f:
				text = f.read()
		else:
			text = XMLTools.inner_text(node)

		text = text.strip("\n")
		text = textwrap.dedent(text)
		text = text.strip("\n")
		text = text.expandtabs(4)

		return cls.handle_text(text, rendered_presentation, node)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import hashlib
import subprocess
from pyradium.Exceptions import InvalidBooleanValueException

class XMLTools():
	@classmethod
	def child_tagname(cls, node, tag_name):
		if node is None:
			return None
		if isinstance(tag_name, str):
			for child in node.childNodes:
				if (child.nodeType == node.ELEMENT_NODE) and (child.tagName == tag_name):
					return child
			return None
		else:
			current = node
			for path_element in tag_name:
				current = cls.child_tagname(current, path_element)
			return current

	@classmethod
	def remove_node(cls, node):
		if node.parentNode is not None:
			node.parentNode.removeChild(node)

	@classmethod
	def replace_node(cls, node, replacement):
		parent = node.parentNode
		if not isinstance(replacement, list):
			parent.replaceChild(replacement, node)
		else:
			result = node.parentNode.replaceChild(replacement[-1], node)
			last_node = replacement[-1]
			for sibling_node in reversed(replacement[ : -1]):
				parent.insertBefore(sibling_node, last_node)
				last_node = sibling_node

	@classmethod
	def remove_siblings_after(cls, node):
		while node is not None:
			next_node = node.nextSibling
			cls.remove_node(node)
			node = next_node

	@classmethod
	def inner_text(cls, node):
		result = [ ]
		def callback(node):
			result.append(node.wholeText)
		cls.walk(node, callback, predicate = lambda node: node.nodeType in [ node.TEXT_NODE, node.CDATA_SECTION_NODE ])
		return "".join(result)

	@classmethod
	def get_ns_tag(cls, node):
		if ":" in node.tagName:
			return node.tagName.split(":")
		else:
			return (None, node.tagName)

	@classmethod
	def walk(cls, node, callback, predicate = None):
		if (predicate is None) or predicate(node):
			callback(node)
		for child in node.childNodes:
			cls.walk(child, callback, predicate)

	@classmethod
	def walk_elements(cls, node, callback):
		cls.walk(node, callback, predicate = lambda node: node.nodeType == node.ELEMENT_NODE)

	@classmethod
	def findall_recurse_predicate(cls, root_node, predicate):
		result = [ ]
		def callback(node):
			if predicate(node):
				result.append(node)
		cls.walk_elements(root_node, callback)
		return result

	@classmethod
	def findall_recurse(cls, root_node, name):
		return cls.findall_recurse_predicate(root_node, predicate = lambda node: (node.nodeType == node.ELEMENT_NODE) and (node.tagName == name))

	@classmethod
	def findall(cls, root_node, name, namespace_uri = "*"):
		result = [ ]
		for node in root_node.childNodes:
			if (node.nodeType == node.ELEMENT_NODE) and (node.tagName == name) and ((namespace_uri == "*") or (node.namespaceURI == namespace_uri)):
				result.append(node)
		return result

	@classmethod
	def normalize_ns(cls, node, known_namespaces = None):
		if known_namespaces is None:
			known_namespaces = { }

		# First collect all known namespace URIs
		present_namespaces = set()
		def visit(node):
			if node.namespaceURI is not None:
				present_namespaces.add(node.namespaceURI)
		cls.walk_elements(node, visit)

		# Then assign them
		assigned_namespaces = { }
		namespace_id = 0
		for uri in sorted(present_namespaces):
			if uri in known_namespaces:
				assigned_namespaces[uri] = known_namespaces[uri]
			else:
				assigned_namespaces[uri] = "ns%d" % (namespace_id)
				namespace_id += 1

		# Remove old namespace assignments
		remove_attributes = set(key for key in node.attributes.keys() if key.startswith("xmlns:"))
		for key in remove_attributes:
			node.attributes.removeNamedItem(key)

		# Append new namespace assignments to root element
		for (uri, key) in assigned_namespaces.items():
			node.setAttribute("xmlns:%s" % (key), uri)

		# Finally, traverse the tree again and change tag names
		def visit(node):
			(ns, tag) = cls.get_ns_tag(node)
			if node.namespaceURI is None:
				new_prefix = None
				new_tagname = tag
			else:
				new_prefix = assigned_namespaces[node.namespaceURI]
				new_tagname = new_prefix + ":" + tag
			node.prefix = new_prefix
			node.tagName = new_tagname
			node.nodeName = new_tagname
		cls.walk_elements(node, visit)

		return assigned_namespaces

	@classmethod
	def inner_toxml(cls, node):
		return "".join(child.toxml() for child in node.childNodes)

	@classmethod
	def clone(cls, node):
		return node.cloneNode(deep = True)

	@classmethod
	def has_sub_elements(cls, node):
		for node in node.childNodes:
			if node.nodeType == node.ELEMENT_NODE:
				return True
		return False

	@classmethod
	def xml_to_dict(cls, node):
		if node is None:
			return None
		if cls.has_sub_elements(node):
			result = { }
			for child in node.childNodes:
				if child.nodeType == node.ELEMENT_NODE:
					key = child.tagName
					value = cls.xml_to_dict(child)
					result[key] = value
			return result
		else:
			return cls.inner_text(node)

	@classmethod
	def get_bool_attr(cls, node, attr_name, default_value = False):
		if not node.hasAttribute(attr_name):
			return default_value
		value = node.getAttribute(attr_name).lower().strip()
		if value in [ "1", "on", "true", "yes" ]:
			return True
		elif value in [ "0", "off", "false", "no" ]:
			return False
		else:
			raise InvalidBooleanValueException("Invalid boolean value for attribute %s of node %s." % (attr_name, node))

class ImageTools():
	@classmethod
	def get_image_info(cls, filename):
		image_info = json.loads(subprocess.check_output([ "convert", filename, "json:-" ]))[0]
		return image_info

class HashTools():
	@classmethod
	def hash_file(cls, filename):
		with open(filename, "rb") as f:
			return hashlib.md5(f.read()).hexdigest()



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import importlib

class BaseController():
	def __init__(self, slide, content_containers, rendered_presentation, options = None):
		self._slide = slide
		self._content_containers = content_containers
		self._rendered_presentation = rendered_presentation
		self._options = options

	@property
	def slide(self):
		return self._slide

	@property
	def content_containers(self):
		return self._content_containers

	@property
	def rendered_presentation(self):
		return self._rendered_presentation

	def render(self):
		return self._slide.emit_content_slide(self._rendered_presentation, self._content_containers)

class ControllerManager():
	def __init__(self, renderer):
		self._renderer = renderer
		self._cached = { }

	def _controller_definition(self, name):
		return self._renderer.template_config.get("controllers", { }).get(name)

	def get_controller(self, slide, content_containers, rendered_presentation):
		definition = self._controller_definition(slide.slide_type)
		options = definition.get("options") if (definition is not None) else None
		if slide.slide_type not in self._cached:
			if definition is not None:
				module_filename = self._renderer.lookup_template_file(definition["file"])
				loader = importlib.machinery.SourceFileLoader("controller_module", module_filename)
				spec = importlib.util.spec_from_loader(loader.name, loader)
				module = importlib.util.module_from_spec(spec)
				loader.exec_module(module)
				handler_class = getattr(module, definition.get("class", "Controller"))
			else:
				handler_class = None
			self._cached[slide.slide_type] = handler_class
		else:
			handler_class = self._cached[slide.slide_type]

		if handler_class is None:
			return None

		controller = handler_class(slide, content_containers, rendered_presentation, options)
		return controller



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
import enum
import collections
from .Exceptions import TimeSpecificationError

class TimeSpecificationType(enum.IntEnum):
	Relative = 0
	Absolute = 1

class TimeSpecification():
	_ABS_REGEX = re.compile(r"((?P<single_value>(\d*\.)?\d+)\s*(?P<unit>min|sec|m|s)|(?P<mins>\d+):(?P<secs>\d+))")
	_REL_REGEX = re.compile(r"(?P<rel_value>(\d*\.)?\d+)")

	def __init__(self, value, spec_type):
		assert(isinstance(spec_type, TimeSpecificationType))
		self._value = value
		self._spec_type = spec_type

	@property
	def value(self):
		return self._value

	@property
	def spec_type(self):
		return self._spec_type

	@classmethod
	def parse(cls, abs_string = None, rel_string = None):
		if (abs_string is None) and (rel_string is None):
			raise TimeSpecificationError("Either absolute or relative timing must be supplied.")
		if (abs_string is not None) and (rel_string is not None):
			raise TimeSpecificationError("Either absolute or relative timing must be supplied, not both.")

		if abs_string is not None:
			match = cls._ABS_REGEX.fullmatch(abs_string)
			if match is None:
				raise TimeSpecificationError("Invalid string for absolute timespecification: \"%s\"" % (abs_string))
			match = match.groupdict()
			if match["unit"] in [ "min", "m" ]:
				value = float(match["single_value"]) * 60
			elif match["unit"] in [ "sec", "s" ]:
				value = float(match["single_value"])
			elif match["unit"] is None:
				value = (60 * int(match["mins"])) + int(match["secs"])
			else:
				raise NotImplementedError(match["unit"])
			return cls(spec_type = TimeSpecificationType.Absolute, value = value)
		else:
			match = cls._REL_REGEX.fullmatch(rel_string)
			if match is None:
				raise TimeSpecificationError("Invalid string for relative timespecification: \"%s\"" % (rel_string))
			match = match.groupdict()
			return cls(spec_type = TimeSpecificationType.Relative, value = float(match["rel_value"]))

	def __str__(self):
		return "TimeSpecification<%s, %.0f>" % (self.spec_type.name, self.value)

class TimeRange():
	_TIME_RE = re.compile(r"\s*(?P<begin_hrs>\d+):(?P<begin_mins>\d{2})\s*-\s*(?P<end_hrs>\d+):(?P<end_mins>\d{2})\s*")
	def __init__(self, range_begin, range_end):
		self._range_begin = range_begin
		self._range_end = range_end

	@property
	def range_begin(self):
		return self._range_begin

	@property
	def range_end(self):
		return self._range_end

	@property
	def begin_time(self):
		return divmod(self.range_begin % 1440, 60)

	@property
	def end_time(self):
		return divmod(self.range_end % 1440, 60)

	@property
	def duration_mins(self):
		return self.range_end - self.range_begin

	@classmethod
	def parse(cls, text):
		match = cls._TIME_RE.fullmatch(text)
		if match is None:
			raise TimeSpecificationError("Invalid time range: \"%s\"" % (text))
		match = match.groupdict()
		match = { key: int(value) for (key, value) in match.items() }
		if not (0 <= match["begin_hrs"] <= 23):
			raise TimeSpecificationError("Invalid time range: \"%s\"; begin hours must be in range 0..23" % (text))
		if not (0 <= match["end_hrs"] <= 23):
			raise TimeSpecificationError("Invalid time range: \"%s\"; end hours must be in range 0..23" % (text))
		if not (0 <= match["begin_mins"] <= 59):
			raise TimeSpecificationError("Invalid time range: \"%s\"; begin minutes must be in range 0..59" % (text))
		if not (0 <= match["end_mins"] <= 59):
			raise TimeSpecificationError("Invalid time range: \"%s\"; end minutes must be in range 0..59" % (text))

		range_begin = (int(match["begin_hrs"]) * 60) + int(match["begin_mins"])
		range_end = (int(match["end_hrs"]) * 60) + int(match["end_mins"])
		if range_end < range_begin:
			range_end += 1440
		return cls(range_begin = range_begin, range_end = range_end)

	def __str__(self):
		return "%d:%02d-%d:%02d" % (self.begin_time[0], self.begin_time[1], self.end_time[0], self.end_time[1])

class TimeRanges():
	_SPLITTER_RE = re.compile(r"\s+")

	def __init__(self, time_ranges):
		assert(all(isinstance(time_range, TimeRange) for time_range in time_ranges))
		self._time_ranges = tuple(time_ranges)

	@classmethod
	def parse(cls, text):
		return cls(tuple(TimeRange.parse(text_value) for text_value in cls._SPLITTER_RE.split(text)))

	@property
	def duration_mins(self):
		return sum(time_range.duration_mins for time_range in self)

	def __iter__(self):
		return iter(self._time_ranges)

	def __str__(self):
		return " ".join(str(time_range) for time_range in self)

class PresentationSchedule():
	_TimeSlice = collections.namedtuple("TimeSlice", [ "slide_no", "slide_ratio", "begin_ratio", "end_ratio", "time_seconds" ])

	def __init__(self, active_presentation_time_minutes):
		if active_presentation_time_minutes is not None:
			self._active_presentation_time_seconds = 60 * active_presentation_time_minutes
		else:
			self._active_presentation_time_seconds = None
		self._time_specs = { }
		self._timeslices = None
		self._max_slide_no = 0

	def have_slide(self, slide_no):
		self._max_slide_no = max(self._max_slide_no, slide_no)

	def set_slide_no(self, slide_no, time_spec):
		assert(isinstance(time_spec, TimeSpecification))
		self._time_specs[slide_no] = time_spec
		self._max_slide_no = max(self._max_slide_no, slide_no)

	def _slide_range(self):
		return range(1, self._max_slide_no + 1)

	def _slide_no_timespec(self):
		for slide_no in self._slide_range():
			yield (slide_no, self._time_specs.get(slide_no))

	def compute(self):
		if self._timeslices is None:
			abs_sum_secs = 0
			rel_sum_pts = 0
			for (slide_no, time_spec) in self._slide_no_timespec():
				if time_spec is not None:
					if time_spec.spec_type == TimeSpecificationType.Absolute:
						abs_sum_secs += time_spec.value
					elif time_spec.spec_type == TimeSpecificationType.Relative:
						rel_sum_pts += time_spec.value
					else:
						raise NotImplementedError(time_spec.spec_type)
				else:
					rel_sum_pts += 1

			if (abs_sum_secs > 0) and (self._active_presentation_time_seconds is None):
				raise TimeSpecificationError("You are free to not define a presentation length beforehand, but then absolute time references may not be used.")

			if self._active_presentation_time_seconds is not None:
				active_rel_time_secs = self._active_presentation_time_seconds - abs_sum_secs
				if active_rel_time_secs < 0:
					raise TimeSpecificationError("Presentation active time is %.0f seconds, but %.0f seconds already allocated for absolute slides." % (self._active_presentation_time_seconds, abs_sum_secs))
			else:
				# Doesn't matter at all, we're only dealing with relative references
				active_rel_time_secs = 1

			if rel_sum_pts == 0:
				# If they are in sum 0, it does not matter anyways, all the
				# numerators will be zero and this way we save two if-elses to
				# avoid divide-by-zero
				rel_sum_pts += 1

			timeslices = { }
			current_slide_ratio = 0
			for (slide_no, time_spec) in self._slide_no_timespec():
				if time_spec is not None:
					if time_spec.spec_type == TimeSpecificationType.Absolute:
						slide_time_secs = time_spec.value
					elif time_spec.spec_type == TimeSpecificationType.Relative:
						slide_time_secs = time_spec.value / rel_sum_pts * active_rel_time_secs
				else:
					slide_time_secs = 1 / rel_sum_pts * active_rel_time_secs

				if self._active_presentation_time_seconds is not None:
					slide_ratio = slide_time_secs / self._active_presentation_time_seconds
				else:
					slide_ratio = slide_time_secs
				timeslice = self._TimeSlice(slide_no = slide_no, slide_ratio = slide_ratio, begin_ratio = current_slide_ratio, end_ratio = current_slide_ratio + slide_ratio, time_seconds = slide_time_secs)
				current_slide_ratio += slide_ratio
				timeslices[slide_no] = timeslice

			self._timeslices = timeslices
		return self._timeslices

	def __getitem__(self, slide_no):
		if self._timeslices is None:
			return self._TimeSlice(slide_no = slide_no, slide_ratio = 1, begin_ratio = 0, end_ratio = 1, time_seconds = 1)
		else:
			return self._timeslices[slide_no]

	def __iter__(self):
		return iter(self._timeslices.values())



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
import argparse
import pyradium
from .MultiCommand import MultiCommand
from .ActionRender import ActionRender
from .ActionServe import ActionServe
from .ActionAcroSort import ActionAcroSort
from .Enums import PresentationMode, PresentationFeature

def _geometry(text):
	text = text.split("x", maxsplit = 1)
	if len(text) != 2:
		raise argparse.ArgumentTypeError("Not a valid geometry: %s" % (text))
	return (int(text[0]), int(text[1]))

def main():
	mc = MultiCommand(trailing_text = "version: pyradium v%s" % (pyradium.VERSION))

	def genparser(parser):
		parser.add_argument("--image-max-dimension", metavar = "pixels", type = int, default = 1920, help = "When rendering imaages, specifies the maximum dimension they're downsized to. The lower this value, the smaller the output files and the lower the quality. Defaults to %(default)d pixels.")
		parser.add_argument("-I", "--include-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional include directory in which, for example, images are located which are referenced from the presentation. Can be issued multiple times.")
		parser.add_argument("--template-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional template directories in which template style files are located. Can be issued multiple times.")
		parser.add_argument("-t", "--template-style", metavar = "name", default = "antonio", help = "Template style to use. Defaults to %(default)s.")
		parser.add_argument("-g", "--geometry", metavar = "width x height", type = _geometry, default = "1280x720", help = "Slide geometry, in pixels. Defaults to %(default)s.")
		parser.add_argument("-r", "--remove-pauses", action = "store_true", help = "Ignore all pause directives and just render the final slides.")
		parser.add_argument("--collapse-animation", action = "store_true", help = "Do not render animations as multiple slides, just show one complete slide.")
		parser.add_argument("-i", "--index-filename", metavar = "filename", default = "index.html", help = "Gives the name of the presentation index file. Defaults to %(default)s. Useful if you want to render multiple presentations in one subdirectory.")
		parser.add_argument("-m", "--presentation-mode", metavar = "{%s}" % (",".join(enumitem.value for enumitem in PresentationMode)), type = PresentationMode, default = "interactive", help = "Generate this type of presentation. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-F", "--presentation-feature", metavar = "{%s}" % (",".join(enumitem.value for enumitem in PresentationFeature)), type = PresentationFeature, default = [ ], action = "append", help = "Enable a specific presentation feature. Can be one of %(choices)s and can be given multiple times.")
		parser.add_argument("-l", "--re-render-loop", action = "store_true", help = "Stay in a continuous loop, re-rendering the presentation if anything changes.")
		parser.add_argument("--re-render-watch", metavar = "path", action = "append", default = [ ], help = "By default, all include files and the template directory is being watched for changes. This option gives additional files or directories upon change of which the presentation should be re-rendered.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite files in destination directory if they already exist.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified more than once.")
		parser.add_argument("infile", help = "Input XML file of the slide show.")
		parser.add_argument("outdir", help = "Output directory the presentation is put into.")
	mc.register("render", "Render a slide show", genparser, action = ActionRender)

	def genparser(parser):
		parser.add_argument("-b", "--bind-addr", metavar = "addr", type = str, default = "127.0.0.1", help = "Address to bind to. Defaults to %(default)s.")
		parser.add_argument("-p", "--port", metavar = "port", type = int, default = 8123, help = "Port to serve directory under. Defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("dirname", help = "Directory that should be served.")
	mc.register("serve", "Serve a slide show as HTTP", genparser, action = ActionServe)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("acrofile", help = "Acronym database JSON file.")
	mc.register("acrosort", "Sort an acryonym database", genparser, action = ActionAcroSort)

	return mc.run(sys.argv[1:])

if __name__ == "__main__":
	main()



#!/usr/bin/python3
#
#	PrefixMatcher - Match the shortest unambiguous prefix
#	Copyright (C) 2011-2019 Johannes Bauer
#
#	This file is part of pycommon.
#
#	pycommon is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pycommon is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pycommon; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#
#	File UUID f75ad3fb-12d0-4368-b14b-9353df125800

class PrefixMatcher():
	def __init__(self, options):
		self._opts = options

	def matchunique(self, value):
		result = self.match(value)
		if len(result) != 1:
			if len(result) == 0:
				raise Exception("'%s' did not match any options." % (value))
			else:
				raise Exception("'%s' is ambiguous. Please clarify further. Available: %s" % (value, ", ".join(sorted(list(result)))))
		return result[0]

	def match(self, value):
		return [ option for option in self._opts if option.startswith(value) ]

if __name__ == "__main__":
	pm = PrefixMatcher([ "import", "install", "foo" ])

	print(pm.match("i"))
	print(pm.matchunique("i"))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import enum
import zlib
import base64

class ExtendedJSONObjects(enum.Enum):
	UncompressedBytes = "6bbc5f9e-6aba-40f4-878c-1ce5f5f50055"
	ZLibCompressedBytes = "f12d83b4-b2dd-4968-8f14-e063970c66fd"

class ExtendedJSONEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, bytes):
			if len(obj) > 1000:
				# Try to compress first
				compressed = zlib.compress(obj)

				# Evaluate if it was worth it
				saved_size_bytes = len(obj) - len(compressed)
				saved_size_percent = 100 * saved_size_bytes / len(obj)

			if (len(obj) <= 1000) or (saved_size_bytes < 1000) or (saved_size_percent < 1):
				# Save raw, not worth it.
				return {
					"__internal_object__":	ExtendedJSONObjects.UncompressedBytes.value,
					"data":					base64.b64encode(obj).decode("ascii"),
				}
			else:
				# Save zlib compressed
				return {
					"__internal_object__":	ExtendedJSONObjects.ZLibCompressedBytes.value,
					"data":					base64.b64encode(compressed).decode("ascii"),
				}
		return json.JSONEncoder.default(self, obj)

	@classmethod
	def _load_object_hook(cls, obj):
		if isinstance(obj, dict) and ("__internal_object__" in obj):
			uuid = obj["__internal_object__"]
			obj_type = ExtendedJSONObjects(uuid)
			if obj_type == ExtendedJSONObjects.UncompressedBytes:
				return base64.b64decode(obj["data"])
			elif obj_type == ExtendedJSONObjects.ZLibCompressedBytes:
				return zlib.decompress(base64.b64decode(obj["data"]))
			else:
				raise NotImplementedError(obj_type)
		else:
			return obj

	@classmethod
	def dumps(cls, obj, minify = False, sort_keys = False, indent = None):
		separators = (",", ":") if minify else None
		return json.dumps(obj, cls = cls, separators = separators, sort_keys = sort_keys, indent = indent)

	@classmethod
	def dump(cls, obj, f, minify = False, sort_keys = False, indent = None):
		separators = (",", ":") if minify else None
		return json.dump(obj, f, cls = cls, separators = separators, sort_keys = sort_keys, indent = indent)

	@classmethod
	def load(cls, f):
		return json.load(f, object_hook = cls._load_object_hook)



#!/usr/bin/python3
#
#	CmdlineEscape - Escape a command line so that it can be safely put in a bash script
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of pycommon.
#
#	pycommon is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pycommon is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pycommon; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#
#	File UUID df05d56b-c766-41a7-a240-0a8bef0a6064

class CmdlineEscape():
	_ESCAPE_CHARS = set(" \\\"';&*()|")

	def __init__(self, always_exported_env = None):
		self._always_exported_env = set(always_exported_env) if (always_exported_env is not None) else tuple()

	@classmethod
	def _needs_escaping(cls, text):
		for char in text:
			if char in cls._ESCAPE_CHARS:
				return True
		return None

	@classmethod
	def _escape(cls, text):
		if cls._needs_escaping(text):
			return "'%s'" % (text.replace("'", "'\\''"))
		else:
			return text

	def cmdline(self, cmd, env = None):
		if env is None:
			env = { }
		else:
			env = dict(env)
		for varname in self._always_exported_env:
			if (varname in os.environ) and (varname not in env):
				env[varname] = os.environ[varname]

		command = " ".join(self._escape(arg) for arg in cmd)
		if env is None:
			return command
		else:
			env_string = " ".join("%s=%s" % (key, self._escape(value)) for (key, value) in sorted(env.items()))
			return (env_string + " " + command).lstrip()

if __name__ == "__main__":
	cle = CmdlineEscape()
	print(cle.cmdline([ "echo", "hello-there" ]))
	print(cle.cmdline([ "echo", "hello there" ]))
	print(cle.cmdline([ "echo", "hello \" there" ]))
	print(cle.cmdline([ "echo", "hello ' there" ]))
	print(cle.cmdline([ "echo", "hello & there" ]))
	print(cle.cmdline([ "echo", "hello!", "you!" ]))
	print(cle.cmdline([ "echo", "hello!(", "you!" ]))
	print(cle.cmdline([ "echo", "()!&;'\" $foo" ]))
	print(cle.cmdline([ "echo", "\\\\" ]))
	print(cle.cmdline([ "echo", "foo 'bar' moo" ]))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry

@XMLHookRegistry.register_hook
class ImgHook(BaseHook):
	_TAG_NAME = "img"

	@classmethod
	def handle(cls, rendered_presentation, node):
		properties = {
			"src":				rendered_presentation.renderer.lookup_include(node.getAttribute("src")),
			"max_dimension":	rendered_presentation.renderer.rendering_params.image_max_dimension,
		}
		img_renderer = rendered_presentation.renderer.get_custom_renderer("img")
		rendered_image = img_renderer.render(properties)
		local_filename = "imgs/img/%s.%s" % (rendered_image.keyhash, rendered_image.data["extension"])

		replacement_node = node.ownerDocument.createElement("div")
		replacement_node.setAttribute("class", "fillimg")

		img_node = node.ownerDocument.createElement("img")
		img_node.setAttribute("src", local_filename)
		img_node.setAttribute("class", "fill")
		replacement_node.appendChild(img_node)

		rendered_presentation.add_file(local_filename, rendered_image.data["img_data"])
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools

@XMLHookRegistry.register_hook
class TexHook(BaseHook):
	_TAG_NAME = "tex"

	@classmethod
	def handle(cls, rendered_presentation, node):
		properties = {
			"formula":	XMLTools.inner_text(node),
			"long":		XMLTools.get_bool_attr(node, "long"),
		}
		tex_renderer = rendered_presentation.renderer.get_custom_renderer("latex")
		rendered_formula = tex_renderer.render(properties)
		local_filename = "imgs/latex/%s.png" % (rendered_formula.keyhash)

		scale_factor = 0.5
		width_px = round(rendered_formula.data["info"]["width"] * scale_factor)
		baseline_px = round(rendered_formula.data["info"]["baseline"] * scale_factor)

		replacement_node = node.ownerDocument.createElement("img")
		replacement_node.setAttribute("src", local_filename)
		if properties["long"]:
			replacement_node.setAttribute("style", "width: %dpx; margin-top: 5px" % (width_px))
		else:
			replacement_node.setAttribute("style", "width: %dpx; margin-bottom: -%dpx; margin-top: 5px" % (width_px, baseline_px))
		replacement_node.setAttribute("alt", properties["formula"])

		rendered_presentation.add_file(local_filename, rendered_formula.data["png_data"])
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .EmoHook import EmoHook
from .SymbolHook import SymbolHook
from .ArrowHook import ArrowHook
from .QuoteHook import QuoteHook
from .TexHook import TexHook
from .TerminalHook import TerminalHook
from .CodeHook import CodeHook
from .ImgHook import ImgHook
from .AcronymHook import AcronymHook
from .TimeHook import TimeHook
from .ExecHook import ExecHook
from .DebugHook import DebugHook



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import xml.dom.minidom
from .Tools import XMLTools

class SVGStyle():
	def __init__(self, style_dict = None):
		assert((style_dict is None) or isinstance(style_dict, dict))
		self._style = style_dict

	@classmethod
	def from_style_str(cls, style_str):
		style_dict = { }
		for style_item in style_str.split(";"):
			style_item = style_item.strip()
			if style_item == "":
				continue
			style_item = style_item.split(":", maxsplit = 1)
			if len(style_item) == 2:
				(key, value) = style_item
				key = key.strip()
				value = value.strip()
				style_dict[key] = value
		return cls(style_dict)

	@classmethod
	def from_node(cls, node):
		return cls.from_style_str(node.getAttribute("style"))

	@property
	def is_visible(self):
		if "display" in self._style:
			hidden = (self["display"] == "none")
			return not hidden
		return True

	def serialize(self):
		return ";".join("%s:%s" % (key, value) for (key, value) in self._style.items())

	def __setitem__(self, key, value):
		self._style[key] = value

	def __getitem__(self, key):
		return self._style.get(key)

	def __str__(self):
		return "SVGStyle<%s>" % (str(self._style))

class SVGLayer():
	def __init__(self, group_node):
		self._node = group_node

	@property
	def layer_id(self):
		return self._node.getAttribute("id")

	@property
	def is_visible(self):
		return SVGStyle.from_node(self._node).is_visible

	def modify_style(self, callback):
		style = SVGStyle.from_node(self._node)
		callback(style)
		self._node.setAttribute("style", style.serialize())

	def hide(self):
		def _callback(style):
			style["display"] = "none"
		self.modify_style(_callback)

	def show(self):
		def _callback(style):
			style["display"] = "inline"
		self.modify_style(_callback)

class SVGTransformation():
	def __init__(self, svg_filename):
		self._xml = xml.dom.minidom.parse(svg_filename)
		self._svg = self._xml.childNodes[0]
		self._layers = [ ]
		self._layers_by_id = { }
		for group_node in XMLTools.findall(self._svg, "g", namespace_uri = "http://www.w3.org/2000/svg"):
			groupmode = group_node.getAttributeNS("http://www.inkscape.org/namespaces/inkscape", "groupmode")
			if groupmode == "layer":
				# Inkscape layer node
				layer = SVGLayer(group_node)
				self._layers.append(layer)
				self._layers_by_id[layer.layer_id] = layer

	@property
	def visible_layer_ids(self):
		for layer in self._layers:
			if layer.is_visible:
				yield layer.layer_id


	@property
	def layer_ids(self):
		for layer in self._layers:
			yield layer.layer_id

	@property
	def layers(self):
		return iter(self._layers)


	def get_layer(self, layer_id):
		return self._layers_by_id[layer_id]

	def apply(self, transformation_dict):
		if transformation_dict["cmd"] == "show_layer":
			self.get_layer(transformation_dict["layer_id"]).show()
		elif transformation_dict["cmd"] == "hide_layer":
			self.get_layer(transformation_dict["layer_id"]).hide()
		else:
			raise NotImplementedError(transformation_dict["cmd"])

	def apply_all(self, transformation_dicts):
		for transformation_dict in transformation_dicts:
			self.apply(transformation_dict)

	def write(self, f):
		self._xml.writexml(f)

	def write_file(self, filename):
		with open(filename, "w") as f:
			self.write(f)

if __name__ == "__main__":
	svg = SVGTransformation("animation_l2hid.svg")
	print(list(svg.layer_ids))
	layer_ids = list(svg.visible_layer_ids)
	print(layer_ids)
	for layer_id in layer_ids:
		svg.get_layer(layer_id).hide()
	svg.get_layer(layer_ids[0]).show()
	svg.write_file("/tmp/x.svg")



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import XMLHookRegistry
from .Tools import XMLTools
from .BaseDirective import BaseDirective
from .RenderableSlide import RenderableSlide
from .PauseRenderer import PauseRenderer

class RenderSlideDirective(BaseDirective):
	def __init__(self, xmlnode):
		assert(xmlnode.tagName == "slide")
		self._dom = xmlnode
		if not self._dom.hasAttribute("type"):
			self._dom.setAttribute("type", "default")
		self._xml_slide_vars = self._get_xml_slide_vars()
		self._content_containers = { }
		for content_node in XMLTools.findall(self._dom, "s:content"):
			self._content_containers[content_node.getAttribute("name")] = content_node
		if len(self._content_containers) == 0:
			self._content_containers["default"] = self._dom

	@property
	def slide_type(self):
		return self._dom.getAttribute("type")

	def get_xml_slide_var(self, key):
		return self._xml_slide_vars.get(key)

	def _get_xml_slide_vars(self):
		# First search DOM for any variables
		slide_vars = { }
		for node in XMLTools.findall(self._dom, "s:var"):
			(key, value) = (node.getAttribute("name"), node.getAttribute("value"))
			slide_vars[key] = value
			XMLTools.remove_node(node)
		return slide_vars

	def compute_slide_vars(self, rendered_presentation, sub_slide_index):
		slide_vars = dict(self._xml_slide_vars)
		slide_vars.update({
			"current_slide_number":	rendered_presentation.current_slide_number,
			"total_slide_count":	rendered_presentation.total_slide_count,
			"chapter":				rendered_presentation.toc.current_text(0),
			"section":				rendered_presentation.toc.current_text(1),
			"subsection":			rendered_presentation.toc.current_text(2),
			"toc":					rendered_presentation.frozen_toc,
			"sub_slide_index":		sub_slide_index,
			"generate_uid":			lambda: rendered_presentation.next_unique_id,
			"timeslice":			rendered_presentation.schedule[rendered_presentation.current_slide_number],
		})
		return slide_vars

	def emit_nocontent_slide(self, rendered_presentation, additional_slide_var_list = None):
		if additional_slide_var_list is None:
			additional_slide_var_list = [ { } ]
		elif not isinstance(additional_slide_var_list, list):
			additional_slide_var_list = [ additional_slide_var_list ]

		rendered_presentation.advance_slide()
		for (sub_slide_index, additional_slide_vars) in enumerate(additional_slide_var_list):
			slide_vars = self.compute_slide_vars(rendered_presentation, sub_slide_index)
			slide_vars.update(additional_slide_vars)
			yield RenderableSlide(slide_type = self.slide_type, content_containers = None, slide_vars = slide_vars)

	def emit_content_slide(self, rendered_presentation, content_containers, additional_slide_vars = None):
		rendered_presentation.advance_slide()
		paused_containers = PauseRenderer(content_containers, honor_pauses = rendered_presentation.renderer.rendering_params.honor_pauses).render()

		for (sub_slide_index, paused_container) in enumerate(paused_containers):
			for container_node in paused_container.values():
				XMLHookRegistry.mangle(rendered_presentation, container_node)

			slide_vars = self.compute_slide_vars(rendered_presentation, sub_slide_index)
			if additional_slide_vars is not None:
				slide_vars.update(additional_slide_vars)
			yield RenderableSlide(slide_type = self.slide_type, content_containers = paused_container, slide_vars = slide_vars)

	def render(self, rendered_presentation):
		controller = rendered_presentation.renderer.controllers.get_controller(self, self._content_containers, rendered_presentation)
		if controller is None:
			yield from self.emit_content_slide(rendered_presentation, self._content_containers)
		else:
			yield from controller.render()

	def __repr__(self):
		return "RenderSlideDirective<%s>" % (self.slide_type)



import sys
import re
import collections
from .FriendlyArgumentParser import FriendlyArgumentParser
from .Schedule import TimeSpecification

class AgendaParser():
	_LINE_RE = re.compile(r"\s*(?P<relspec>\*)?(?P<timespec>[^\s]+)\s+(?P<text>.*)")
	_Item = collections.namedtuple("Item", [ "rel_value", "abs_value", "text" ])

	def __init__(self, args):
		self._args = args
		self._begin_time = TimeSpecification.parse(abs_string = args.begin_time)
		self._end_time = TimeSpecification.parse(abs_string = args.end_time)
		self._abs_sum = 0
		self._rel_sum = 0
		self._items = [ ]
		self._parse_file(self._args.filename)

	def _parse_line(self, line):
		result = self._LINE_RE.fullmatch(line)
		if not result:
			raise Exception("Cannot parse: '%s'" % (line))
		result = result.groupdict()
		if result["relspec"] is not None:
			rel_value = float(result["timespec"])
			self._rel_sum += rel_value
			item = self._Item(rel_value = rel_value, abs_value = None, text = result["text"])
		else:
			abs_string = result["timespec"]
			if abs_string.isdigit():
				abs_string += " sec"
			abs_value = TimeSpecification.parse(abs_string = abs_string)
			self._abs_sum += abs_value.value
			item = self._Item(rel_value = None, abs_value = abs_value, text = result["text"])
		self._items.append(item)

	def _parse_file(self, filename):
		for (lineno, line) in enumerate(open(filename), 1):
			line = line.rstrip("\r\n")
			if line.startswith("#"):
				continue
			try:
				self._parse_line(line)
			except:
				print("Parse error in line %d:" % (lineno), file = sys.stderr)
				raise

	def dump(self):
		total_time = self._end_time.value - self._begin_time.value
		if total_time <= 0:
			raise Exception("Total time is negative or zero.")

		if self._abs_sum > total_time:
			raise Exception("More absolute time (%d minutes) allocated than in presentation (%d minutes)." % (self._abs_sum, total_time))

		rel_time = total_time - self._abs_sum
		if (rel_time > 0) and (self._rel_sum == 0):
			raise Exception("Time left in presentation (%d minutes) and no relative portion used." % (rel_time))

		rel_time_left = rel_time
		rel_pts_left = self._rel_sum
		current_time = self._begin_time.value
		for item in self._items:
			if item.rel_value is not None:
				if abs(item.rel_value - rel_pts_left) > 0.01:
					this_time = (item.rel_value / self._rel_sum) * rel_time
					this_time = round(this_time / self._args.round_mins) * self._args.round_mins
					rel_pts_left -= item.rel_value
					rel_time_left -= this_time
				else:
					this_time = round(rel_time_left / self._args.round_mins) * self._args.round_mins
			else:
				this_time = item.abs_value.value

			end_time = current_time + this_time
			begin_time_str = "%d:%02d" % (current_time // 60, current_time % 60)
			end_time_str = "%d:%02d" % (end_time // 60, end_time % 60)
			if this_time >= 60:
				duration_str = "%d:%02d" % (this_time // 60, this_time % 60)
			else:
				duration_str = "%d" % (this_time)
			print("\t\t\t<li><b>%s - %s</b>: %s (%s)</li>" % (begin_time_str, end_time_str, item.text, duration_str))

			current_time = end_time

		print(self._abs_sum, self._rel_sum)

if __name__ == "__main__":
	parser = FriendlyArgumentParser(description = "Agenda parser.")
	parser.add_argument("-b", "--begin-time", metavar = "time", default = "0:00", help = "Begin at this timeslot. Defaults to %(default)s.")
	parser.add_argument("-e", "--end-time", metavar = "time", default = "24:00", help = "End at this timeslot. Defaults to %(default)s.")
	parser.add_argument("-o", "--offset-time", metavar = "time", help = "Offset everything by this amount of time.")
	parser.add_argument("-r", "--round-mins", metavar = "minutes", type = int, default = 5, help = "Round to this many minutes. Defaults to %(default)d minutes.")
	parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
	parser.add_argument("filename", help = "File to parse")
	args = parser.parse_args(sys.argv[1:])

	ap = AgendaParser(args)
	ap.dump()



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .Tools import XMLTools
from .Exceptions import DuplicateOrderException

class PauseRenderer():
	def __init__(self, content_containers, honor_pauses = True):
		self._content_containers = content_containers
		self._honor_pauses = honor_pauses
		self._used_order_ids = set()

	def _clone_containers(self):
		return { name: XMLTools.clone(container_node) for (name, container_node) in self._content_containers.items() }

	def _enumerate_pause_nodes_of(self, root_node):
		for pause_node in XMLTools.findall_recurse(root_node, "s:pause"):
			if not pause_node.hasAttribute("order"):
				if len(self._used_order_ids) == 0:
					assigned_order_id = 1
				else:
					assigned_order_id = max(self._used_order_ids) + 1
			else:
				assigned_order_id = int(pause_node.getAttribute("order"))
			if assigned_order_id in self._used_order_ids:
				raise DuplicateOrderException("Duplicate order ID: %d" % (assigned_order_id))
			self._used_order_ids.add(assigned_order_id)
			pause_node.setAttribute("order", str(assigned_order_id))

	def _enumerate_pause_nodes(self):
		for (name, container_node) in sorted(self._content_containers.items()):
			self._enumerate_pause_nodes_of(container_node)

	def _debug_result(self, rendered_containers):
		if len(rendered_containers) <= 1:
			return
		print("%d rendered containers:" % (len(rendered_containers)))
		for (index, rendered_container) in enumerate(rendered_containers):
			for (name, container_node) in sorted(rendered_container.items()):
				print("%s.%d:" % (name, index))
				print(XMLTools.inner_toxml(container_node))
				print()
		print("=" * 120)

	def render(self):
		if self._honor_pauses:
			self._enumerate_pause_nodes()

			rendered_containers = [ self._clone_containers() for _ in range(len(self._used_order_ids) + 1) ]
			sorted_order_ids = list(sorted(self._used_order_ids))

			for (max_order_id, rendered_container) in zip(sorted_order_ids, rendered_containers):
				for container_node in rendered_container.values():
					for pause_node in XMLTools.findall_recurse(container_node, "s:pause"):
						order = int(pause_node.getAttribute("order"))
						if order >= max_order_id:
							XMLTools.remove_siblings_after(pause_node)
						else:
							XMLTools.remove_node(pause_node)
		else:
			rendered_containers = [ self._clone_containers() ]

		# Remove all pause nodes from last container (the one which renders the
		# whole slide)
		for container_node in rendered_containers[-1].values():
			for pause_node in XMLTools.findall_recurse(container_node, "s:pause"):
				XMLTools.remove_node(pause_node)

#		self._debug_result(rendered_containers)
		return rendered_containers



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
import subprocess
from pyradium.Tools import ImageTools, HashTools
from pyradium.RendererCache import BaseRenderer
from pyradium.Exceptions import UsageException
from pyradium.SVGTransformation import SVGTransformation

class ImageRenderer(BaseRenderer):
	def __init__(self):
		super().__init__()

	@property
	def name(self):
		return "img"

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def _render_raw_svg(self, src, max_dimension, svg_transform = None):
		with tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = ".png") as output_file:
			svg_info = ImageTools.get_image_info(src)
			if svg_info["image"]["geometry"]["width"] > svg_info["image"]["geometry"]["height"]:
				scale_param = "-w"
			else:
				scale_param = "-h"
			cmd = [ "inkscape", "-o", output_file.name, scale_param, str(max_dimension), src ]
			subprocess.check_call(cmd)
			extension = "png"
			img_data = output_file.read()
		return (extension, img_data)

	def _render_svg(self, src, max_dimension, svg_transform = None):
		if svg_transform is None:
			return self._render_raw_svg(src, max_dimension)
		else:
			# Transform SVG before rendering it
			svg = SVGTransformation(src)
			svg.apply_all(svg_transform)
			with tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = ".svg", mode = "w") as transformed_svg:
				svg.write(transformed_svg)
				transformed_svg.flush()
				return self._render_raw_svg(transformed_svg.name, max_dimension)

	def _render_raster_bitmap(self, src, max_dimension):
		# Resize using ImageMagick
		if src.lower().endswith(".jpg") or src.lower().endswith(".jpeg"):
			extension = "jpg"
		else:
			extension = "png"

		with tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = "." + extension) as output_file:
			cmd = [ "convert", "-geometry", ">%dx%d" % (max_dimension, max_dimension), src, output_file.name ]
			subprocess.check_call(cmd)
			img_data = output_file.read()
		return (extension, img_data)

	def rendering_key(self, property_dict):
		src = property_dict["src"]
		return {
			"srchash":		HashTools.hash_file(src),
		}

	def render(self, property_dict):
		src = property_dict["src"]
		max_dimension = property_dict["max_dimension"]

		if src.lower().endswith(".svg"):
			(extension, img_data) = self._render_svg(src, max_dimension, svg_transform = property_dict.get("svg_transform"))
		else:
			if "svg_transform" in property_dict:
				raise UsageException("SVG transformation requested, but source file is not in SVG format: %s / %s" % (src, str(property_dict)))
			(extension, img_data) = self._render_raster_bitmap(src, max_dimension)

		image = {
			"extension":	extension,
			"img_data":		img_data,
		}
		return image

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(ImageRenderer())
	print(renderer.render({
		"max_dimension": 100,
		"src":	"test.svg",
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Schedule import TimeSpecification

@XMLHookRegistry.register_hook
class TimeHook(BaseHook):
	_TAG_NAME = "time"

	@classmethod
	def handle(cls, rendered_presentation, node):
		abs_string = node.getAttribute("abs") if node.hasAttribute("abs") else None
		rel_string = node.getAttribute("rel") if node.hasAttribute("rel") else None
		time_spec = TimeSpecification.parse(abs_string = abs_string, rel_string = rel_string)
		rendered_presentation.schedule.set_slide_no(rendered_presentation.current_slide_number, time_spec)
		return None



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import collections

ResolvedAcronym = collections.namedtuple("ResolvedAcronym", [ "acronym_id", "acronym", "text", "uri" ])

class AcronymDirective():
	def __init__(self, node):
		self._src = node.getAttribute("src")

	def render(self, rendered_presentation):
		filename = rendered_presentation.renderer.lookup_include(self._src)
		acronyms = rendered_presentation.renderer.get_custom_renderer("acronym")
		acronyms.load_database(filename)

class Acronyms():
	def __init__(self):
		self._acronyms = { }
		self._loaded_files = set()
		self._looked_up = set()
		self._unresolvable = set()

	def get_all_used_acronyms(self):
		used_acronyms = [ self.resolve(acronym_id) for acronym_id in self._looked_up ]
		used_acronyms.sort()
		return used_acronyms

	def load_database(self, filename):
		if filename in self._loaded_files:
			return
		with open(filename) as f:
			acronyms = json.load(f)
		for (acronym_id, acronym_data) in acronyms.items():
			resolved_acronym = ResolvedAcronym(acronym_id = acronym_id, acronym = acronym_data.get("ac", acronym_id), text = acronym_data["text"], uri = acronym_data.get("uri"))
			self._acronyms[acronym_id] = resolved_acronym

	def resolve(self, acronym_id):
		if acronym_id in self._acronyms:
			self._looked_up.add(acronym_id)
		else:
			if acronym_id not in self._unresolvable:
				print("Warning: Acronym '%s' not in acronym database." % (acronym_id))
				self._unresolvable.add(acronym_id)
		return self._acronyms.get(acronym_id)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class PyRadiumException(Exception): pass

class CallingProcessException(PyRadiumException): pass

class MalformedXMLInputException(PyRadiumException): pass

class SlideException(PyRadiumException): pass
class UndefinedContentException(SlideException): pass
class TemplateErrorException(SlideException): pass
class TimeSpecificationError(SlideException): pass
class UsageException(SlideException): pass

class XMLHookRegistryException(PyRadiumException): pass

class DuplicateOrderException(PyRadiumException): pass
class InvalidBooleanValueException(PyRadiumException): pass
class FailedToLookupFileException(PyRadiumException): pass

class FailedToExecuteSubprocessException(PyRadiumException): pass



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import tempfile
import shutil
import contextlib
from .BaseAction import BaseAction

class ActionAcroSort(BaseAction):
	def run(self):
		with open(self._args.acrofile) as f:
			acronyms = json.load(f)

		rows = [ ]
		for (acro_id, acro_data) in acronyms.items():
			rows.append((acro_id, acro_data["text"], acro_data.get("acronym")))

		lengths = [  max(len(row[colid]) for row in rows if row[colid] is not None) for colid in range(len(rows[0])) ]

		allowed_keys = set([ "text", "acronym", "uri" ])
		lines = [ ]
		for (acro_id, acro_data) in sorted(acronyms.items()):
			illegal_keys = set(acro_data.keys()) - allowed_keys
			if len(illegal_keys) > 0:
				print("Refusing to process because of illegal key(s): %s" % (", ".join(sorted(illegal_keys))))
				return 1

			line = [ ]
			line.append("%-*s" % (lengths[0] + 3, "\"%s\":" % (acro_id)))
			line.append(" { ")
			line.append("%-*s" % (lengths[1] + 12, "\"text\": \"%s\", " % (acro_data["text"])))
			if "acronym" in acro_data:
				inner = "\"acronym\": \"%s\", " % (acro_data["acronym"])
			else:
				inner = ""
			line.append("%-*s" % (lengths[2] + 15, inner))

			if "uri" in acro_data:
				line.append("\"uri\": \"%s\", " % (acro_data["uri"]))


			line = "".join(line)
			line = line.rstrip(", ")
			line = line + " }"
			lines.append(line)
		jsondata = "{\n\t" + (",\n\t".join(lines)) + "\n}"

		with contextlib.suppress(FileNotFoundError):
			with tempfile.NamedTemporaryFile(mode = "w") as f:
				f.write(jsondata)
				f.flush()
				shutil.move(f.name, self._args.acrofile)

		return 0



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import string
import collections

class TOCCommand(enum.IntEnum):
	NestingIncrease = 0
	NestingDecrease = 1
	Item = 2

class TOCTranscription(enum.IntEnum):
	Integer = 0
	LowerAlpha = 1
	UpperAlpha = 2

class _TOCInstructionOpcode(enum.IntEnum):
	TOCItem = 0
	CounterSet = 1
	CounterAdd = 2
	TranscriptionUpdate = 3
	SeparatorUpdate = 4

_TOCItem = collections.namedtuple("TOCItem", [ "level", "text", "pages" ])
_NormalizedTOCItem = collections.namedtuple("NormalizedTOCItem", [ "depth", "text", "pages" ])
_TOCInstruction = collections.namedtuple("TOCInstruction", [ "opcode", "data" ])

class FrozenTOC():
	TOCEntry = collections.namedtuple("TOCEntry", [ "order", "index", "depth", "local_number", "full_number", "text", "full_text", "pages" ])

	def __init__(self, toc):
		self._toc = toc
		self._levels = self._determine_levels(toc)
		self._instructions = self._normalize_instructions(toc)
		self._entries = [ ]
		self._index_by_full_number = { }

		self._text = collections.defaultdict(str)
		self._counter = collections.defaultdict(int)
		self._transcriptions = { }
		self._separators = { }
		self._unroll()

	@property
	def max_depth(self):
		return self.level_count

	@property
	def toc_item_count(self):
		return len(self._entries)

	@property
	def level_count(self):
		return len(self._levels)

	@property
	def toc(self):
		return self._toc

	def _seperator(self, depth):
		return self._separators.get(depth, ".")

	def _determine_levels(self, toc):
		levels = set()
		for instruction in toc.instructions:
			if instruction.opcode == _TOCInstructionOpcode.TOCItem:
				levels.add(instruction.data.level)
		return tuple(sorted(levels))

	@staticmethod
	def _transcribe_alpha(index, alphabet):
		letters = [ ]
		while index > 0:
			if len(letters) > 0:
				index -= 1
			(index, char) = divmod(index, len(alphabet))
			letters.append(alphabet[char])
		if len(letters) == 0:
			letters.append(alphabet[0])
		return "".join(reversed(letters))

	def _transcribe(self, depth, value):
		transcriptor = self._transcriptions.get(depth, TOCTranscription.Integer)
		if transcriptor == TOCTranscription.Integer:
			return str(value)
		elif transcriptor == TOCTranscription.LowerAlpha:
			return self._transcribe_alpha(value - 1, string.ascii_lowercase)
		elif transcriptor == TOCTranscription.UpperAlpha:
			return self._transcribe_alpha(value - 1, string.ascii_uppercase)
		else:
			raise NotImplementedError(transcriptor)

	def _normalize_instructions(self, toc):
		depth_by_level = { level: depth for (depth, level) in enumerate(self._levels, 1) }

		normalized_instructions = [ ]
		for item in toc.instructions:
			if item.opcode == _TOCInstructionOpcode.TOCItem:
				normalized_instructions.append(_TOCInstruction(opcode = item.opcode, data = _NormalizedTOCItem(depth = depth_by_level[item.data.level], text = item.data.text, pages = item.data.pages)))
			else:
				normalized_instructions.append(_TOCInstruction(opcode = item.opcode, data = { depth_by_level[level]: data for (level, data) in item.data.items() if (level in depth_by_level) }))
		return normalized_instructions

	def _create_entry(self, depth, pages):
		order = tuple(depth for (level, depth) in sorted(self._counter.items()))

		local_number = self._transcribe(depth, self._counter[depth]) + self._seperator(depth)

		full_number = [ ]
		for i in range(1, depth + 1):
			full_number.append(self._transcribe(i, self._counter[i]))
			if i != depth:
				full_number.append(self._seperator(i))
		full_number = "".join(full_number)

		full_text = [ ]
		for i in range(1, depth + 1):
			full_text.append(self._text[i])

		entry_index = len(self._entries)
		if full_number not in self._index_by_full_number:
			self._index_by_full_number[full_number] = entry_index

		return self.TOCEntry(order = order, index = entry_index, depth = depth, local_number = local_number, full_number = full_number, text = self._text[depth], full_text = full_text, pages = pages)

	def _unroll(self):
		current_depth = -1
		for instruction in self._instructions:
			if instruction.opcode == _TOCInstructionOpcode.TranscriptionUpdate:
				self._transcriptions.update(instruction.data)
			elif instruction.opcode == 	_TOCInstructionOpcode.CounterSet:
				self._counter.update(instruction.data)
			elif instruction.opcode == 	_TOCInstructionOpcode.CounterAdd:
				for (key, value) in instruction.data.items():
					self._counter[key] += value
			elif instruction.opcode == 	_TOCInstructionOpcode.TOCItem:
				depth = instruction.data.depth

				if depth != current_depth:
					for i in range(depth + 1, self.max_depth + 1):
						self._counter[i] = 0
						self._text[i] = ""

				self._counter[depth] += 1
				self._text[depth] = instruction.data.text
				self._entries.append(self._create_entry(depth, instruction.data.pages))

				current_depth = depth
			else:
				raise NotImplementedError(instruction.opcode)


	def emit_commands(self, toc_items):
		current_depth = 0
		for entry in toc_items:
			if entry.depth > current_depth:
				for i in range(entry.depth - current_depth):
					yield (TOCCommand.NestingIncrease, None)
			elif entry.depth < current_depth:
				for i in range(current_depth - entry.depth):
					yield (TOCCommand.NestingDecrease, None)
			current_depth = entry.depth
			yield (TOCCommand.Item, entry)

		for i in range(current_depth):
			yield (TOCCommand.NestingDecrease, None)

	def subset(self, start_at = None, end_before = None, max_items = None):
		start_index = 0
		if start_at is not None:
			if isinstance(start_at, str):
				start_index = self._index_by_full_number[start_at]
			else:
				start_index = start_at
		seen = 0
		current_depth = 0

		for entry in self._entries[start_index : ]:
			if (end_before is not None) and (entry.full_number == end_before):
				break

			seen += 1
			if (max_items is not None) and (seen > max_items):
				break

			yield entry

	def count_toc_items(self, start_at = None, end_before = None):
		return len(list(self.subset(start_at = start_at, end_before = end_before)))

	def __iter__(self):
		return self.emit_commands(self._entries)

	def dump(self):
		for entry in self._entries:
			print(entry)

class GenericTOC():
	"""\
	'levels' are arbitrary IDs, not necessarily consecutive
		e.g. chapter is level 4
		section is level 9
		subsection is level 10

	They are mapped to 'depths', which are index 1. In the aforementioned example
		chapter is level 4 but depth 1
		section is level 9 but depth 2
		subsection is level 10 but depth 3
	"""
	def __init__(self):
		self._instructions = [ ]
		self._text = { }
		self._last_toc_item = { }

	def current_text(self, level):
		return self._text.get(level)

	@property
	def instructions(self):
		return iter(self._instructions)

	def set_transcription(self, level, transcription):
		assert(isinstance(transcription, TOCTranscription))
		self._instructions.append(_TOCInstruction(opcode = _TOCInstructionOpcode.TranscriptionUpdate, data = { level: transcription }))

	def set_counter(self, level, value):
		self._instructions.append(_TOCInstruction(opcode = _TOCInstructionOpcode.CounterSet, data = { level: value }))

	def add_counter(self, level, value):
		self._instructions.append(_TOCInstruction(opcode = _TOCInstructionOpcode.CounterAdd, data = { level: value }))

	def new_heading(self, level, text, at_page = None):
		self._text[level] = text
		toc_item = _TOCItem(level = level, text = text, pages = set())
		self._instructions.append(_TOCInstruction(opcode = _TOCInstructionOpcode.TOCItem, data = toc_item))
		self._last_toc_item[level] = toc_item

	def at_page(self, page_no):
		for (level, toc_item) in self._last_toc_item.items():
			toc_item.pages.add(page_no)

	def finalize(self):
		return FrozenTOC(self)

	def dump(self):
		for instruction in self._instructions:
			if instruction.opcode == _TOCInstructionOpcode.TOCItem:
				print(instruction.data)

if __name__ == "__main__":
	toc = GenericTOC()
	toc.set_transcription(6, TOCTranscription.LowerAlpha)
	toc.new_heading(0, "Intro")
	toc.new_heading(1, "Background")
	toc.new_heading(1, "Related Work")
	toc.new_heading(1, "Things")
	toc.new_heading(0, "Main")
	toc.new_heading(1, "Apparatus")
	toc.new_heading(5, "Design")
	toc.new_heading(5, "Implementation")
	for x in range(3):
		toc.new_heading(6, "fooo %d" % (x))
	toc.new_heading(1, "Setup")
	toc.new_heading(5, "Initital")
	toc.new_heading(5, "Final")
	toc.new_heading(0, "Results")
	toc.set_counter(0, 0)
	toc.set_transcription(0, TOCTranscription.UpperAlpha)
	toc.new_heading(0, "Data Set")
	toc.new_heading(1, "Foobar")
	toc.new_heading(1, "Barfoo")
	toc.new_heading(1, "Mookoo")

	frozen_toc = toc.finalize()
	for cmd in frozen_toc:
		print(cmd)

	print("-" * 120)
	for cmd in frozen_toc.subset(start_at = "2.1", end_before = "3", max_items = 4):
		print(cmd)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import unittest
from pyradium.CmdlineParser import CmdlineParser, TrailingEscapeCharException, InvalidEscapeCharException

class CmdlineParserTests(unittest.TestCase):
	def test_end_meta(self):
		with self.assertRaises(TrailingEscapeCharException):
			CmdlineParser().parse("foo bar\\")
		CmdlineParser().parse("foo bar\\\\")

	def test_invalid_escape(self):
		with self.assertRaises(InvalidEscapeCharException):
			CmdlineParser().parse("foo \\x foo")

	def test_valid_escape(self):
		CmdlineParser().parse("foo \\\\ bar")
		CmdlineParser().parse("foo \\\\ bar\\\\")
		CmdlineParser().parse("foo \\  bar")
		CmdlineParser().parse("foo \\\"  bar")

	def test_simple_parse(self):
		self.assertEqual(CmdlineParser().parse("foo bar moo"), [ "foo", "bar", "moo" ])
		self.assertEqual(CmdlineParser().parse("foo bar   moo"), [ "foo", "bar", "moo" ])
		self.assertEqual(CmdlineParser().parse("foo          bar   moo"), [ "foo", "bar", "moo" ])

	def test_trailing_space(self):
		self.assertEqual(CmdlineParser().parse("foo bar moo "), [ "foo", "bar", "moo" ])
		self.assertEqual(CmdlineParser().parse("foo bar moo            "), [ "foo", "bar", "moo" ])

	def test_leading_space(self):
		self.assertEqual(CmdlineParser().parse(" foo bar moo"), [ "foo", "bar", "moo" ])
		self.assertEqual(CmdlineParser().parse("            foo bar moo"), [ "foo", "bar", "moo" ])

	def test_quotation_1(self):
		self.assertEqual(CmdlineParser().parse("foo \"bar koo\" moo"), [ "foo", "bar koo", "moo" ])
		self.assertEqual(CmdlineParser().parse("foo \"bar ' koo\" moo"), [ "foo", "bar ' koo", "moo" ])

	def test_quotation_2(self):
		self.assertEqual(CmdlineParser().parse("foo 'bar koo' moo"), [ "foo", "bar koo", "moo" ])
		self.assertEqual(CmdlineParser().parse("foo 'bar \" koo' moo"), [ "foo", "bar \" koo", "moo" ])

	def test_quotation_3(self):
		self.assertEqual(CmdlineParser().parse("foo ''bar\"\"\"moo\"'koo' moo"), [ "foo", "barmookoo", "moo" ])
